//
// Copyright 2021 Vulcanize, Inc.
//

import assert from 'assert';
import { DeepPartial, FindConditions, FindManyOptions, ObjectLiteral } from 'typeorm';
import debug from 'debug';
import { ethers, constants, providers } from 'ethers';
import { GraphQLResolveInfo } from 'graphql';

import { JsonFragment } from '@ethersproject/abi';
import { MappingKey, StorageLayout } from '@cerc-io/solidity-mapper';
import {
  Indexer as BaseIndexer,
  IndexerInterface,
  ValueResult,
  ServerConfig,
  JobQueue,
  Where,
  QueryOptions,
  BlockHeight,
  ResultMeta,
  updateSubgraphState,
  dumpSubgraphState,
  GraphWatcherInterface,
  StateKind,
  StateStatus,
  ResultEvent,
  getResultEvent,
  DatabaseInterface,
  Clients,
  EthClient,
  UpstreamConfig,
  EthFullBlock,
  EthFullTransaction,
  ExtraEventData
} from '@cerc-io/util';
import { initClients } from '@cerc-io/cli';
import { GraphWatcher } from '@cerc-io/graph-node';

import KeeperRegistryArtifacts from './artifacts/KeeperRegistry.json';
import OrchestratorArtifacts from './artifacts/Orchestrator.json';
import StrategyRegistryArtifacts from './artifacts/StrategyRegistry.json';
import VaultRegistryArtifacts from './artifacts/VaultRegistry.json';
import BundleRegistryArtifacts from './artifacts/BundleRegistry.json';
import GasVaultArtifacts from './artifacts/GasVault.json';
import WhitelistRegistryArtifacts from './artifacts/WhitelistRegistry.json';
import VaultArtifacts from './artifacts/Vault.json';
import DynamicJobArtifacts from './artifacts/DynamicJob.json';
import { Database, ENTITIES, SUBGRAPH_ENTITIES } from './database';
import { createInitialState, handleEvent, createStateDiff, createStateCheckpoint } from './hooks';
import { Contract } from './entity/Contract';
import { Event } from './entity/Event';
import { SyncStatus } from './entity/SyncStatus';
import { StateSyncStatus } from './entity/StateSyncStatus';
import { BlockProgress } from './entity/BlockProgress';
import { State } from './entity/State';
/* eslint-disable @typescript-eslint/no-unused-vars */
import { Keeper } from './entity/Keeper';
import { Runner } from './entity/Runner';
import { QueueTimeline } from './entity/QueueTimeline';
import { PermissionUpdate } from './entity/PermissionUpdate';
import { OrchestratorAction } from './entity/OrchestratorAction';
import { Vote } from './entity/Vote';
import { ActionFailure } from './entity/ActionFailure';
import { Creator } from './entity/Creator';
import { CreatorWithdrawal } from './entity/CreatorWithdrawal';
import { Strategy } from './entity/Strategy';
import { Vault } from './entity/Vault';
import { VaultSnapshot } from './entity/VaultSnapshot';
import { PrevVaultSnapshot } from './entity/PrevVaultSnapshot';
import { PrevDailyVaultSnapshot } from './entity/PrevDailyVaultSnapshot';
import { PrevMonthlyVaultSnapshot } from './entity/PrevMonthlyVaultSnapshot';
import { PrevAnnualVaultSnapshot } from './entity/PrevAnnualVaultSnapshot';
import { PrevWeeklyVaultSnapshot } from './entity/PrevWeeklyVaultSnapshot';
import { VaultStateChanged } from './entity/VaultStateChanged';
import { VaultDeposit } from './entity/VaultDeposit';
import { VaultWithdraw } from './entity/VaultWithdraw';
import { Depositor } from './entity/Depositor';
import { VaultBeacon } from './entity/VaultBeacon';
import { VaultPosition } from './entity/VaultPosition';
import { LiquiditySteer } from './entity/LiquiditySteer';
import { OrchestratorReward } from './entity/OrchestratorReward';
import { OrchestratorRewardTracker } from './entity/OrchestratorRewardTracker';
import { OrchestratorRewardSnapshot } from './entity/OrchestratorRewardSnapshot';
import { Bundle } from './entity/Bundle';
import { Job } from './entity/Job';
import { JobExecution } from './entity/JobExecution';
import { JobGasDeposited } from './entity/JobGasDeposited';
import { JobGasWithdrawn } from './entity/JobGasWithdrawn';
import { VaultGasUsed } from './entity/VaultGasUsed';
import { VaultGasDeposited } from './entity/VaultGasDeposited';
import { VaultGasWithdrawn } from './entity/VaultGasWithdrawn';
import { WhiteListManager } from './entity/WhiteListManager';
import { WhiteListVaultPermission } from './entity/WhiteListVaultPermission';
import { LeaderBoard } from './entity/LeaderBoard';
import { VaultTransfer } from './entity/VaultTransfer';
/* eslint-enable @typescript-eslint/no-unused-vars */

import { FrothyEntity } from './entity/FrothyEntity';

// eslint-disable-next-line @typescript-eslint/no-unused-vars
const log = debug('vulcanize:indexer');

const KIND_KEEPERREGISTRY = 'KeeperRegistry';

const KIND_ORCHESTRATOR = 'Orchestrator';

const KIND_STRATEGYREGISTRY = 'StrategyRegistry';

const KIND_VAULTREGISTRY = 'VaultRegistry';

const KIND_BUNDLEREGISTRY = 'BundleRegistry';

const KIND_GASVAULT = 'GasVault';

const KIND_WHITELISTREGISTRY = 'WhitelistRegistry';

const KIND_VAULT = 'Vault';

const KIND_DYNAMICJOB = 'DynamicJob';

export class Indexer implements IndexerInterface {
  _db: Database;
  _ethClient: EthClient;
  _ethProvider: providers.JsonRpcProvider;
  _baseIndexer: BaseIndexer;
  _serverConfig: ServerConfig;
  _upstreamConfig: UpstreamConfig;
  _graphWatcher: GraphWatcher;

  _abiMap: Map<string, JsonFragment[]>;
  _storageLayoutMap: Map<string, StorageLayout>;
  _contractMap: Map<string, ethers.utils.Interface>;
  eventSignaturesMap: Map<string, string[]>;

  _entityTypesMap: Map<string, { [key: string]: string }>;
  _relationsMap: Map<any, { [key: string]: any }>;

  _subgraphStateMap: Map<string, any>;

  constructor (
    config: {
      server: ServerConfig;
      upstream: UpstreamConfig;
    },
    db: DatabaseInterface,
    clients: Clients,
    ethProvider: providers.JsonRpcProvider,
    jobQueue: JobQueue,
    graphWatcher?: GraphWatcherInterface
  ) {
    assert(db);
    assert(clients.ethClient);

    this._db = db as Database;
    this._ethClient = clients.ethClient;
    this._ethProvider = ethProvider;
    this._serverConfig = config.server;
    this._upstreamConfig = config.upstream;
    this._baseIndexer = new BaseIndexer(config, this._db, this._ethClient, this._ethProvider, jobQueue);
    assert(graphWatcher);
    this._graphWatcher = graphWatcher as GraphWatcher;

    this._abiMap = new Map();
    this._storageLayoutMap = new Map();
    this._contractMap = new Map();
    this.eventSignaturesMap = new Map();

    const { abi: KeeperRegistryABI } = KeeperRegistryArtifacts;

    const { abi: OrchestratorABI } = OrchestratorArtifacts;

    const { abi: StrategyRegistryABI } = StrategyRegistryArtifacts;

    const { abi: VaultRegistryABI } = VaultRegistryArtifacts;

    const { abi: BundleRegistryABI } = BundleRegistryArtifacts;

    const { abi: GasVaultABI } = GasVaultArtifacts;

    const { abi: WhitelistRegistryABI } = WhitelistRegistryArtifacts;

    const { abi: VaultABI } = VaultArtifacts;

    const { abi: DynamicJobABI } = DynamicJobArtifacts;

    assert(KeeperRegistryABI);
    this._abiMap.set(KIND_KEEPERREGISTRY, KeeperRegistryABI);

    const KeeperRegistryContractInterface = new ethers.utils.Interface(KeeperRegistryABI);
    this._contractMap.set(KIND_KEEPERREGISTRY, KeeperRegistryContractInterface);

    const KeeperRegistryEventSignatures = Object.values(KeeperRegistryContractInterface.events).map(value => {
      return KeeperRegistryContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_KEEPERREGISTRY, KeeperRegistryEventSignatures);

    assert(OrchestratorABI);
    this._abiMap.set(KIND_ORCHESTRATOR, OrchestratorABI);

    const OrchestratorContractInterface = new ethers.utils.Interface(OrchestratorABI);
    this._contractMap.set(KIND_ORCHESTRATOR, OrchestratorContractInterface);

    const OrchestratorEventSignatures = Object.values(OrchestratorContractInterface.events).map(value => {
      return OrchestratorContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_ORCHESTRATOR, OrchestratorEventSignatures);

    assert(StrategyRegistryABI);
    this._abiMap.set(KIND_STRATEGYREGISTRY, StrategyRegistryABI);

    const StrategyRegistryContractInterface = new ethers.utils.Interface(StrategyRegistryABI);
    this._contractMap.set(KIND_STRATEGYREGISTRY, StrategyRegistryContractInterface);

    const StrategyRegistryEventSignatures = Object.values(StrategyRegistryContractInterface.events).map(value => {
      return StrategyRegistryContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_STRATEGYREGISTRY, StrategyRegistryEventSignatures);

    assert(VaultRegistryABI);
    this._abiMap.set(KIND_VAULTREGISTRY, VaultRegistryABI);

    const VaultRegistryContractInterface = new ethers.utils.Interface(VaultRegistryABI);
    this._contractMap.set(KIND_VAULTREGISTRY, VaultRegistryContractInterface);

    const VaultRegistryEventSignatures = Object.values(VaultRegistryContractInterface.events).map(value => {
      return VaultRegistryContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_VAULTREGISTRY, VaultRegistryEventSignatures);

    assert(BundleRegistryABI);
    this._abiMap.set(KIND_BUNDLEREGISTRY, BundleRegistryABI);

    const BundleRegistryContractInterface = new ethers.utils.Interface(BundleRegistryABI);
    this._contractMap.set(KIND_BUNDLEREGISTRY, BundleRegistryContractInterface);

    const BundleRegistryEventSignatures = Object.values(BundleRegistryContractInterface.events).map(value => {
      return BundleRegistryContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_BUNDLEREGISTRY, BundleRegistryEventSignatures);

    assert(GasVaultABI);
    this._abiMap.set(KIND_GASVAULT, GasVaultABI);

    const GasVaultContractInterface = new ethers.utils.Interface(GasVaultABI);
    this._contractMap.set(KIND_GASVAULT, GasVaultContractInterface);

    const GasVaultEventSignatures = Object.values(GasVaultContractInterface.events).map(value => {
      return GasVaultContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_GASVAULT, GasVaultEventSignatures);

    assert(WhitelistRegistryABI);
    this._abiMap.set(KIND_WHITELISTREGISTRY, WhitelistRegistryABI);

    const WhitelistRegistryContractInterface = new ethers.utils.Interface(WhitelistRegistryABI);
    this._contractMap.set(KIND_WHITELISTREGISTRY, WhitelistRegistryContractInterface);

    const WhitelistRegistryEventSignatures = Object.values(WhitelistRegistryContractInterface.events).map(value => {
      return WhitelistRegistryContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_WHITELISTREGISTRY, WhitelistRegistryEventSignatures);

    assert(VaultABI);
    this._abiMap.set(KIND_VAULT, VaultABI);

    const VaultContractInterface = new ethers.utils.Interface(VaultABI);
    this._contractMap.set(KIND_VAULT, VaultContractInterface);

    const VaultEventSignatures = Object.values(VaultContractInterface.events).map(value => {
      return VaultContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_VAULT, VaultEventSignatures);

    assert(DynamicJobABI);
    this._abiMap.set(KIND_DYNAMICJOB, DynamicJobABI);

    const DynamicJobContractInterface = new ethers.utils.Interface(DynamicJobABI);
    this._contractMap.set(KIND_DYNAMICJOB, DynamicJobContractInterface);

    const DynamicJobEventSignatures = Object.values(DynamicJobContractInterface.events).map(value => {
      return DynamicJobContractInterface.getEventTopic(value);
    });
    this.eventSignaturesMap.set(KIND_DYNAMICJOB, DynamicJobEventSignatures);

    this._entityTypesMap = new Map();
    this._populateEntityTypesMap();

    this._relationsMap = new Map();
    this._populateRelationsMap();

    this._subgraphStateMap = new Map();
  }

  get serverConfig (): ServerConfig {
    return this._serverConfig;
  }

  get upstreamConfig (): UpstreamConfig {
    return this._upstreamConfig;
  }

  get storageLayoutMap (): Map<string, StorageLayout> {
    return this._storageLayoutMap;
  }

  get contractMap (): Map<string, ethers.utils.Interface> {
    return this._contractMap;
  }

  get graphWatcher (): GraphWatcher {
    return this._graphWatcher;
  }

  async init (): Promise<void> {
    await this._baseIndexer.fetchContracts();
    await this._baseIndexer.fetchStateStatus();
  }

  async switchClients (): Promise<void> {
    const { ethClient, ethProvider } = await this._baseIndexer.switchClients(initClients);
    this._ethClient = ethClient;
    this._ethProvider = ethProvider;
    this._graphWatcher.switchClients({ ethClient, ethProvider });
  }

  async isGetLogsRequestsSlow (): Promise<boolean> {
    return this._baseIndexer.isGetLogsRequestsSlow();
  }

  async getMetaData (block: BlockHeight): Promise<ResultMeta | null> {
    return this._baseIndexer.getMetaData(block);
  }

  getResultEvent (event: Event): ResultEvent {
    return getResultEvent(event);
  }

  async getStorageValue (storageLayout: StorageLayout, blockHash: string, contractAddress: string, variable: string, ...mappingKeys: MappingKey[]): Promise<ValueResult> {
    return this._baseIndexer.getStorageValue(
      storageLayout,
      blockHash,
      contractAddress,
      variable,
      ...mappingKeys
    );
  }

  async getEntitiesForBlock (blockHash: string, tableName: string): Promise<any[]> {
    return this._db.getEntitiesForBlock(blockHash, tableName);
  }

  async processInitialState (contractAddress: string, blockHash: string): Promise<any> {
    // Call initial state hook.
    return createInitialState(this, contractAddress, blockHash);
  }

  async processStateCheckpoint (contractAddress: string, blockHash: string): Promise<boolean> {
    // Call checkpoint hook.
    return createStateCheckpoint(this, contractAddress, blockHash);
  }

  async processCanonicalBlock (blockHash: string, blockNumber: number): Promise<void> {
    console.time('time:indexer#processCanonicalBlock-finalize_auto_diffs');
    // Finalize staged diff blocks if any.
    await this._baseIndexer.finalizeDiffStaged(blockHash);
    console.timeEnd('time:indexer#processCanonicalBlock-finalize_auto_diffs');

    // Call custom stateDiff hook.
    await createStateDiff(this, blockHash);

    this._graphWatcher.pruneEntityCacheFrothyBlocks(blockHash, blockNumber);
  }

  async processCheckpoint (blockHash: string): Promise<void> {
    // Return if checkpointInterval is <= 0.
    const checkpointInterval = this._serverConfig.checkpointInterval;
    if (checkpointInterval <= 0) return;

    console.time('time:indexer#processCheckpoint-checkpoint');
    await this._baseIndexer.processCheckpoint(this, blockHash, checkpointInterval);
    console.timeEnd('time:indexer#processCheckpoint-checkpoint');
  }

  async processCLICheckpoint (contractAddress: string, blockHash?: string): Promise<string | undefined> {
    return this._baseIndexer.processCLICheckpoint(this, contractAddress, blockHash);
  }

  async getPrevState (blockHash: string, contractAddress: string, kind?: string): Promise<State | undefined> {
    return this._db.getPrevState(blockHash, contractAddress, kind);
  }

  async getLatestState (contractAddress: string, kind: StateKind | null, blockNumber?: number): Promise<State | undefined> {
    return this._db.getLatestState(contractAddress, kind, blockNumber);
  }

  async getStatesByHash (blockHash: string): Promise<State[]> {
    return this._baseIndexer.getStatesByHash(blockHash);
  }

  async getStateByCID (cid: string): Promise<State | undefined> {
    return this._baseIndexer.getStateByCID(cid);
  }

  async getStates (where: FindConditions<State>): Promise<State[]> {
    return this._db.getStates(where);
  }

  getStateData (state: State): any {
    return this._baseIndexer.getStateData(state);
  }

  // Method used to create auto diffs (diff_staged).
  async createDiffStaged (contractAddress: string, blockHash: string, data: any): Promise<void> {
    console.time('time:indexer#createDiffStaged-auto_diff');
    await this._baseIndexer.createDiffStaged(contractAddress, blockHash, data);
    console.timeEnd('time:indexer#createDiffStaged-auto_diff');
  }

  // Method to be used by createStateDiff hook.
  async createDiff (contractAddress: string, blockHash: string, data: any): Promise<void> {
    const block = await this.getBlockProgress(blockHash);
    assert(block);

    await this._baseIndexer.createDiff(contractAddress, block, data);
  }

  // Method to be used by createStateCheckpoint hook.
  async createStateCheckpoint (contractAddress: string, blockHash: string, data: any): Promise<void> {
    const block = await this.getBlockProgress(blockHash);
    assert(block);

    return this._baseIndexer.createStateCheckpoint(contractAddress, block, data);
  }

  // Method to be used by export-state CLI.
  async createCheckpoint (contractAddress: string, blockHash: string): Promise<string | undefined> {
    const block = await this.getBlockProgress(blockHash);
    assert(block);

    return this._baseIndexer.createCheckpoint(this, contractAddress, block);
  }

  // Method to be used by fill-state CLI.
  async createInit (blockHash: string, blockNumber: number): Promise<void> {
    // Create initial state for contracts.
    await this._baseIndexer.createInit(this, blockHash, blockNumber);
  }

  async saveOrUpdateState (state: State): Promise<State> {
    return this._baseIndexer.saveOrUpdateState(state);
  }

  async removeStates (blockNumber: number, kind: StateKind): Promise<void> {
    await this._baseIndexer.removeStates(blockNumber, kind);
  }

  async getSubgraphEntity<Entity extends ObjectLiteral> (
    entity: new () => Entity,
    id: string,
    block: BlockHeight,
    queryInfo: GraphQLResolveInfo
  ): Promise<any> {
    const data = await this._graphWatcher.getEntity(entity, id, this._relationsMap, block, queryInfo);

    return data;
  }

  async getSubgraphEntities<Entity extends ObjectLiteral> (
    entity: new () => Entity,
    block: BlockHeight,
    where: { [key: string]: any } = {},
    queryOptions: QueryOptions = {},
    queryInfo: GraphQLResolveInfo
  ): Promise<any[]> {
    return this._graphWatcher.getEntities(entity, this._relationsMap, block, where, queryOptions, queryInfo);
  }

  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  async triggerIndexingOnEvent (event: Event, extraData: ExtraEventData): Promise<void> {
    const resultEvent = this.getResultEvent(event);

    console.time('time:indexer#processEvent-mapping_code');
    // Call subgraph handler for event.
    await this._graphWatcher.handleEvent(resultEvent, extraData);
    console.timeEnd('time:indexer#processEvent-mapping_code');

    // Call custom hook function for indexing on event.
    await handleEvent(this, resultEvent);
  }

  async processEvent (event: Event, extraData: ExtraEventData): Promise<void> {
    // Trigger indexing of data based on the event.
    await this.triggerIndexingOnEvent(event, extraData);
  }

  async processBlock (blockProgress: BlockProgress): Promise<void> {
    console.time('time:indexer#processBlock-init_state');
    // Call a function to create initial state for contracts.
    await this._baseIndexer.createInit(this, blockProgress.blockHash, blockProgress.blockNumber);
    console.timeEnd('time:indexer#processBlock-init_state');

    this._graphWatcher.updateEntityCacheFrothyBlocks(blockProgress);
  }

  async processBlockAfterEvents (blockHash: string, blockNumber: number, extraData: ExtraEventData): Promise<void> {
    console.time('time:indexer#processBlockAfterEvents-mapping_code');
    // Call subgraph handler for block.
    await this._graphWatcher.handleBlock(blockHash, blockNumber, extraData);
    console.timeEnd('time:indexer#processBlockAfterEvents-mapping_code');

    console.time('time:indexer#processBlockAfterEvents-dump_subgraph_state');
    // Persist subgraph state to the DB.
    await this.dumpSubgraphState(blockHash);
    console.timeEnd('time:indexer#processBlockAfterEvents-dump_subgraph_state');
  }

  parseEventNameAndArgs (watchedContracts: Contract[], logObj: any): { eventParsed: boolean, eventDetails: any } {
    const { topics, data } = logObj;
    let logDescription: ethers.utils.LogDescription | undefined;

    for (const watchedContract of watchedContracts) {
      const contract = this._contractMap.get(watchedContract.kind);
      assert(contract);

      try {
        logDescription = contract.parseLog({ data, topics });
        break;
      } catch (err) {
        // Continue loop only if no matching event found
        if (!((err as Error).message.includes('no matching event'))) {
          throw err;
        }
      }
    }

    if (!logDescription) {
      return { eventParsed: false, eventDetails: {} };
    }

    const { eventName, eventInfo, eventSignature } = this._baseIndexer.parseEvent(logDescription);

    return {
      eventParsed: true,
      eventDetails: {
        eventName,
        eventInfo,
        eventSignature
      }
    };
  }

  async getStateSyncStatus (): Promise<StateSyncStatus | undefined> {
    return this._db.getStateSyncStatus();
  }

  async updateStateSyncStatusIndexedBlock (blockNumber: number, force?: boolean): Promise<StateSyncStatus | undefined> {
    if (!this._serverConfig.enableState) {
      return;
    }

    const dbTx = await this._db.createTransactionRunner();
    let res;

    try {
      res = await this._db.updateStateSyncStatusIndexedBlock(dbTx, blockNumber, force);
      await dbTx.commitTransaction();
    } catch (error) {
      await dbTx.rollbackTransaction();
      throw error;
    } finally {
      await dbTx.release();
    }

    return res;
  }

  async updateStateSyncStatusCheckpointBlock (blockNumber: number, force?: boolean): Promise<StateSyncStatus> {
    const dbTx = await this._db.createTransactionRunner();
    let res;

    try {
      res = await this._db.updateStateSyncStatusCheckpointBlock(dbTx, blockNumber, force);
      await dbTx.commitTransaction();
    } catch (error) {
      await dbTx.rollbackTransaction();
      throw error;
    } finally {
      await dbTx.release();
    }

    return res;
  }

  async getLatestCanonicalBlock (): Promise<BlockProgress | undefined> {
    const syncStatus = await this.getSyncStatus();
    assert(syncStatus);

    if (syncStatus.latestCanonicalBlockHash === constants.HashZero) {
      return;
    }

    const latestCanonicalBlock = await this.getBlockProgress(syncStatus.latestCanonicalBlockHash);
    assert(latestCanonicalBlock);

    return latestCanonicalBlock;
  }

  async getLatestStateIndexedBlock (): Promise<BlockProgress> {
    return this._baseIndexer.getLatestStateIndexedBlock();
  }

  async addContracts (): Promise<void> {
    // Watching all the contracts in the subgraph.
    await this._graphWatcher.addContracts();
  }

  async watchContract (address: string, kind: string, checkpoint: boolean, startingBlock: number, context?: any): Promise<void> {
    return this._baseIndexer.watchContract(address, kind, checkpoint, startingBlock, context);
  }

  updateStateStatusMap (address: string, stateStatus: StateStatus): void {
    this._baseIndexer.updateStateStatusMap(address, stateStatus);
  }

  cacheContract (contract: Contract): void {
    return this._baseIndexer.cacheContract(contract);
  }

  async saveEventEntity (dbEvent: Event): Promise<Event> {
    return this._baseIndexer.saveEventEntity(dbEvent);
  }

  async saveEvents (dbEvents: Event[]): Promise<void> {
    return this._baseIndexer.saveEvents(dbEvents);
  }

  async getEventsByFilter (blockHash: string, contract?: string, name?: string): Promise<Array<Event>> {
    return this._baseIndexer.getEventsByFilter(blockHash, contract, name);
  }

  isContractAddressWatched (address : string): Contract[] | undefined {
    return this._baseIndexer.isContractAddressWatched(address);
  }

  getWatchedContracts (): Contract[] {
    return this._baseIndexer.getWatchedContracts();
  }

  getContractsByKind (kind: string): Contract[] {
    return this._baseIndexer.getContractsByKind(kind);
  }

  async getProcessedBlockCountForRange (fromBlockNumber: number, toBlockNumber: number): Promise<{ expected: number, actual: number }> {
    return this._baseIndexer.getProcessedBlockCountForRange(fromBlockNumber, toBlockNumber);
  }

  async getEventsInRange (fromBlockNumber: number, toBlockNumber: number): Promise<Array<Event>> {
    return this._baseIndexer.getEventsInRange(fromBlockNumber, toBlockNumber, this._serverConfig.gql.maxEventsBlockRange);
  }

  async getEvents (options: FindManyOptions<Event>): Promise<Array<Event>> {
    return this._db.getEvents(options);
  }

  async getSyncStatus (): Promise<SyncStatus | undefined> {
    return this._baseIndexer.getSyncStatus();
  }

  async getBlocks (blockFilter: { blockHash?: string, blockNumber?: number }): Promise<any> {
    return this._baseIndexer.getBlocks(blockFilter);
  }

  async getBlockByHash (blockHash?: string): Promise<{ block: any }> {
    return this._baseIndexer.getBlockByHash(blockHash);
  }

  async updateSyncStatusIndexedBlock (blockHash: string, blockNumber: number, force = false): Promise<SyncStatus> {
    return this._baseIndexer.updateSyncStatusIndexedBlock(blockHash, blockNumber, force);
  }

  async updateSyncStatusChainHead (blockHash: string, blockNumber: number, force = false): Promise<SyncStatus> {
    return this._baseIndexer.updateSyncStatusChainHead(blockHash, blockNumber, force);
  }

  async updateSyncStatusCanonicalBlock (blockHash: string, blockNumber: number, force = false): Promise<SyncStatus> {
    const syncStatus = this._baseIndexer.updateSyncStatusCanonicalBlock(blockHash, blockNumber, force);
    await this.pruneFrothyEntities(blockNumber);

    return syncStatus;
  }

  async updateSyncStatusProcessedBlock (blockHash: string, blockNumber: number, force = false): Promise<SyncStatus> {
    return this._baseIndexer.updateSyncStatusProcessedBlock(blockHash, blockNumber, force);
  }

  async updateSyncStatusIndexingError (hasIndexingError: boolean): Promise<SyncStatus | undefined> {
    return this._baseIndexer.updateSyncStatusIndexingError(hasIndexingError);
  }

  async updateSyncStatus (syncStatus: DeepPartial<SyncStatus>): Promise<SyncStatus> {
    return this._baseIndexer.updateSyncStatus(syncStatus);
  }

  async getEvent (id: string): Promise<Event | undefined> {
    return this._baseIndexer.getEvent(id);
  }

  async getBlockProgress (blockHash: string): Promise<BlockProgress | undefined> {
    return this._baseIndexer.getBlockProgress(blockHash);
  }

  async getBlockProgressEntities (where: FindConditions<BlockProgress>, options: FindManyOptions<BlockProgress>): Promise<BlockProgress[]> {
    return this._baseIndexer.getBlockProgressEntities(where, options);
  }

  async getBlocksAtHeight (height: number, isPruned: boolean): Promise<BlockProgress[]> {
    return this._baseIndexer.getBlocksAtHeight(height, isPruned);
  }

  async fetchAndSaveFilteredEventsAndBlocks (startBlock: number, endBlock: number): Promise<{
    blockProgress: BlockProgress,
    events: DeepPartial<Event>[],
    ethFullBlock: EthFullBlock;
    ethFullTransactions: EthFullTransaction[];
  }[]> {
    return this._baseIndexer.fetchAndSaveFilteredEventsAndBlocks(startBlock, endBlock, this.eventSignaturesMap, this.parseEventNameAndArgs.bind(this));
  }

  async fetchEventsForContracts (blockHash: string, blockNumber: number, addresses: string[]): Promise<DeepPartial<Event>[]> {
    return this._baseIndexer.fetchEventsForContracts(blockHash, blockNumber, addresses, this.eventSignaturesMap, this.parseEventNameAndArgs.bind(this));
  }

  async saveBlockAndFetchEvents (block: DeepPartial<BlockProgress>): Promise<[
    BlockProgress,
    DeepPartial<Event>[],
    EthFullTransaction[]
  ]> {
    return this._saveBlockAndFetchEvents(block);
  }

  async getBlockEvents (blockHash: string, where: Where, queryOptions: QueryOptions): Promise<Array<Event>> {
    return this._baseIndexer.getBlockEvents(blockHash, where, queryOptions);
  }

  async removeUnknownEvents (block: BlockProgress): Promise<void> {
    return this._baseIndexer.removeUnknownEvents(Event, block);
  }

  async markBlocksAsPruned (blocks: BlockProgress[]): Promise<void> {
    await this._baseIndexer.markBlocksAsPruned(blocks);

    await this._graphWatcher.pruneEntities(FrothyEntity, blocks, SUBGRAPH_ENTITIES);
  }

  async pruneFrothyEntities (blockNumber: number): Promise<void> {
    await this._graphWatcher.pruneFrothyEntities(FrothyEntity, blockNumber);
  }

  async resetLatestEntities (blockNumber: number): Promise<void> {
    await this._graphWatcher.resetLatestEntities(blockNumber);
  }

  async updateBlockProgress (block: BlockProgress, lastProcessedEventIndex: number): Promise<BlockProgress> {
    return this._baseIndexer.updateBlockProgress(block, lastProcessedEventIndex);
  }

  async getAncestorAtHeight (blockHash: string, height: number): Promise<string> {
    return this._baseIndexer.getAncestorAtHeight(blockHash, height);
  }

  async resetWatcherToBlock (blockNumber: number): Promise<void> {
    const entities = [...ENTITIES, FrothyEntity];
    await this._baseIndexer.resetWatcherToBlock(blockNumber, entities);

    await this.resetLatestEntities(blockNumber);
  }

  async clearProcessedBlockData (block: BlockProgress): Promise<void> {
    const entities = [...ENTITIES, FrothyEntity];
    await this._baseIndexer.clearProcessedBlockData(block, entities);

    await this.resetLatestEntities(block.blockNumber);
  }

  getEntityTypesMap (): Map<string, { [key: string]: string }> {
    return this._entityTypesMap;
  }

  getRelationsMap (): Map<any, { [key: string]: any }> {
    return this._relationsMap;
  }

  updateSubgraphState (contractAddress: string, data: any): void {
    return updateSubgraphState(this._subgraphStateMap, contractAddress, data);
  }

  async dumpSubgraphState (blockHash: string, isStateFinalized = false): Promise<void> {
    return dumpSubgraphState(this, this._subgraphStateMap, blockHash, isStateFinalized);
  }

  _populateEntityTypesMap (): void {
    this._entityTypesMap.set('Keeper', {
      id: 'ID',
      bondHeld: 'BigDecimal',
      index: 'BigInt',
      status: 'String'
    });
    this._entityTypesMap.set('Runner', {
      id: 'ID',
      bondHeld: 'BigInt',
      createdAt: 'BigInt'
    });
    this._entityTypesMap.set('QueueTimeline', {
      id: 'ID',
      keeper: 'Keeper',
      timeDelay: 'BigInt',
      queued: 'Boolean'
    });
    this._entityTypesMap.set('PermissionUpdate', {
      id: 'ID',
      keeper: 'Keeper',
      timeStamp: 'BigInt',
      action: 'String'
    });
    this._entityTypesMap.set('OrchestratorAction', {
      id: 'ID',
      timestamp: 'BigInt',
      from: 'Keeper',
      lastUpdated: 'BigInt',
      state: 'String',
      status: 'String',
      recipients: 'String',
      vault: 'String',
      transactionHash: 'String',
      hash: 'String',
      gasUsed: 'BigInt'
    });
    this._entityTypesMap.set('Vote', {
      id: 'ID',
      timestamp: 'BigInt',
      by: 'Keeper',
      action: 'OrchestratorAction'
    });
    this._entityTypesMap.set('ActionFailure', {
      id: 'ID',
      timeStamp: 'BigInt',
      method: 'BigInt',
      action: 'OrchestratorAction'
    });
    this._entityTypesMap.set('Creator', {
      id: 'ID',
      revenue: 'BigDecimal',
      totalValueLocked: 'BigDecimal',
      totalYield: 'BigDecimal'
    });
    this._entityTypesMap.set('CreatorWithdrawal', {
      id: 'ID',
      timeStamp: 'BigInt',
      creator: 'Creator',
      amount: 'BigDecimal'
    });
    this._entityTypesMap.set('Strategy', {
      id: 'ID',
      name: 'String',
      createdAt: 'BigInt',
      admin: 'String',
      creator: 'Creator',
      executionBundle: 'String'
    });
    this._entityTypesMap.set('Vault', {
      id: 'ID',
      deployer: 'String',
      token0: 'String',
      token1: 'String',
      totalAmount0: 'BigDecimal',
      totalAmount1: 'BigDecimal',
      createdAt: 'BigInt',
      lastSnapshot: 'BigInt',
      pool: 'String',
      state: 'BigInt',
      strategyToken: 'Strategy',
      annualPercentageYield: 'BigDecimal',
      annualPercentageDailyYield: 'BigDecimal',
      annualPercentageMonthlyYield: 'BigDecimal',
      annualPercentageYearlyYield: 'BigDecimal',
      lastTotalT0ValuePerLPT: 'BigDecimal',
      accruedStrategistFees0: 'BigInt',
      accruedStrategistFees1: 'BigInt',
      fees0: 'BigInt',
      fees1: 'BigInt',
      beaconName: 'String',
      gasUsed: 'BigInt',
      gasDeposited: 'BigInt',
      totalLPTokensIssued: 'BigInt',
      token1Symbol: 'String',
      decimals: 'BigInt',
      feeTier: 'BigInt',
      name: 'String',
      symbol: 'String',
      token0Balance: 'BigInt',
      token0Decimals: 'BigInt',
      token0Name: 'String',
      token0Symbol: 'String',
      token1Balance: 'BigInt',
      token1Decimals: 'BigInt',
      token1Name: 'String',
      payloadIpfs: 'String',
      vaultManager: 'String',
      averageFeeArrPerSecond: 'BigDecimal',
      totalSnapshots: 'BigInt',
      annualFeeARR: 'BigDecimal',
      dailyFeeAPR: 'BigDecimal',
      weeklyFeeAPR: 'BigDecimal',
      totalValueLockedToken0: 'BigDecimal',
      totalValueLockedToken1: 'BigDecimal'
    });
    this._entityTypesMap.set('VaultSnapshot', {
      id: 'ID',
      timestamp: 'BigInt',
      vaultAddress: 'Vault',
      totalAmount0: 'BigInt',
      totalAmount1: 'BigInt',
      sqrtPriceX96: 'BigInt',
      totalSupply: 'BigInt',
      fees1: 'BigInt',
      fees0: 'BigInt',
      annualFeeAPR: 'BigDecimal',
      dailyFeeAPR: 'BigDecimal',
      weeklyFeeAPR: 'BigDecimal',
      transactionHash: 'String'
    });
    this._entityTypesMap.set('PrevVaultSnapshot', {
      id: 'ID',
      timestamp: 'BigInt',
      vaultAddress: 'Vault',
      totalAmount0: 'BigInt',
      totalAmount1: 'BigInt',
      sqrtPriceX96: 'BigInt',
      totalSupply: 'BigInt',
      fees1: 'BigInt',
      fees0: 'BigInt'
    });
    this._entityTypesMap.set('PrevDailyVaultSnapshot', {
      id: 'ID',
      timestamp: 'BigInt',
      vaultAddress: 'Vault',
      totalAmount0: 'BigInt',
      totalAmount1: 'BigInt',
      sqrtPriceX96: 'BigInt',
      totalSupply: 'BigInt',
      fees1: 'BigInt',
      fees0: 'BigInt'
    });
    this._entityTypesMap.set('PrevMonthlyVaultSnapshot', {
      id: 'ID',
      timestamp: 'BigInt',
      vaultAddress: 'Vault',
      totalAmount0: 'BigInt',
      totalAmount1: 'BigInt',
      sqrtPriceX96: 'BigInt',
      totalSupply: 'BigInt',
      fees1: 'BigInt',
      fees0: 'BigInt'
    });
    this._entityTypesMap.set('PrevAnnualVaultSnapshot', {
      id: 'ID',
      timestamp: 'BigInt',
      vaultAddress: 'Vault',
      totalAmount0: 'BigInt',
      totalAmount1: 'BigInt',
      sqrtPriceX96: 'BigInt',
      totalSupply: 'BigInt',
      fees1: 'BigInt',
      fees0: 'BigInt'
    });
    this._entityTypesMap.set('PrevWeeklyVaultSnapshot', {
      id: 'ID',
      timestamp: 'BigInt',
      vaultAddress: 'Vault',
      totalAmount0: 'BigInt',
      totalAmount1: 'BigInt',
      sqrtPriceX96: 'BigInt',
      totalSupply: 'BigInt',
      fees1: 'BigInt',
      fees0: 'BigInt',
      averageFeeArrPerSecond: 'BigDecimal',
      totalSnapshots: 'BigInt',
      weeklyFeeApr: 'BigDecimal'
    });
    this._entityTypesMap.set('VaultStateChanged', {
      id: 'ID',
      timeStamp: 'BigInt',
      vault: 'Vault',
      status: 'BigInt'
    });
    this._entityTypesMap.set('VaultDeposit', {
      id: 'ID',
      vault: 'Vault',
      amount0: 'BigDecimal',
      amount1: 'BigDecimal',
      token0: 'String',
      token1: 'String',
      sender: 'String',
      transactionHash: 'String',
      timeStamp: 'BigInt',
      shares: 'BigInt',
      executor: 'String',
      depositCaller: 'String',
      blockNumber: 'BigInt'
    });
    this._entityTypesMap.set('VaultWithdraw', {
      id: 'ID',
      vault: 'Vault',
      amount0: 'BigDecimal',
      amount1: 'BigDecimal',
      token0: 'String',
      token1: 'String',
      timeStamp: 'BigInt',
      shares: 'BigInt',
      sender: 'String',
      transactionHash: 'String',
      blockNumber: 'BigInt'
    });
    this._entityTypesMap.set('Depositor', {
      id: 'ID',
      vault: 'Vault',
      account: 'String',
      executor: 'String',
      depositCaller: 'String',
      depositedAmount0: 'BigInt',
      depositedAmount1: 'BigInt',
      withdrawnAmount0: 'BigInt',
      withdrawnAmount1: 'BigInt',
      liquidityAmount0: 'BigInt',
      liquidityAmount1: 'BigInt',
      createdTimestamp: 'BigInt',
      updatedTimestamp: 'BigInt',
      shares: 'BigInt'
    });
    this._entityTypesMap.set('VaultBeacon', {
      id: 'ID',
      address: 'String',
      ipfsHash: 'String',
      name: 'String',
      status: 'String',
      timestamp: 'BigInt',
      updateTimestamp: 'BigInt'
    });
    this._entityTypesMap.set('VaultPosition', {
      id: 'ID',
      upperTick: 'BigInt',
      lowerTick: 'BigInt',
      vault: 'Vault',
      relativeWeight: 'BigInt',
      timestamp: 'BigInt'
    });
    this._entityTypesMap.set('LiquiditySteer', {
      id: 'ID',
      vault: 'Vault',
      timeStamp: 'BigInt',
      tick: 'BigInt'
    });
    this._entityTypesMap.set('OrchestratorReward', {
      id: 'ID',
      timeStamp: 'BigInt',
      updatedTimeStamp: 'BigInt',
      address: 'String',
      reward: 'BigInt',
      trackerId: 'OrchestratorRewardTracker'
    });
    this._entityTypesMap.set('OrchestratorRewardTracker', {
      id: 'ID',
      timeStamp: 'BigInt'
    });
    this._entityTypesMap.set('OrchestratorRewardSnapshot', {
      id: 'ID',
      timeStamp: 'BigInt',
      updatedTimeStamp: 'BigInt',
      address: 'String',
      reward: 'BigInt'
    });
    this._entityTypesMap.set('Bundle', {
      id: 'ID',
      createdAt: 'BigInt',
      bundle: 'String',
      creator: 'String',
      host: 'String',
      source: 'String',
      output: 'String',
      active: 'Boolean',
      infoHash: 'String',
      hash: 'String'
    });
    this._entityTypesMap.set('Job', {
      id: 'ID',
      name: 'String',
      timestamp: 'BigInt',
      jobInfo: 'String',
      targetAddresses: 'String',
      ipfsHash: 'String',
      status: 'BigInt',
      failedCounts: 'BigInt',
      vaultAddress: 'Vault',
      jobHash: 'String',
      gasUsed: 'BigInt'
    });
    this._entityTypesMap.set('JobExecution', {
      id: 'ID',
      timestamp: 'BigInt',
      jobHash: 'String',
      executor: 'String',
      status: 'String',
      jobIdString: 'String',
      jobId: 'Job'
    });
    this._entityTypesMap.set('JobGasDeposited', {
      id: 'ID',
      timestamp: 'BigInt',
      creator: 'String',
      amount: 'BigInt'
    });
    this._entityTypesMap.set('JobGasWithdrawn', {
      id: 'ID',
      timestamp: 'BigInt',
      creator: 'String',
      amount: 'BigInt'
    });
    this._entityTypesMap.set('VaultGasUsed', {
      id: 'ID',
      timestamp: 'BigInt',
      vault: 'String',
      actionHash: 'String',
      amount: 'BigInt'
    });
    this._entityTypesMap.set('VaultGasDeposited', {
      id: 'ID',
      timestamp: 'BigInt',
      vault: 'String',
      origin: 'String',
      amount: 'BigInt'
    });
    this._entityTypesMap.set('VaultGasWithdrawn', {
      id: 'ID',
      timestamp: 'BigInt',
      vault: 'String',
      to: 'String',
      amount: 'BigInt'
    });
    this._entityTypesMap.set('WhiteListManager', {
      id: 'ID',
      address: 'String',
      vault: 'String',
      timestamp: 'BigInt'
    });
    this._entityTypesMap.set('WhiteListVaultPermission', {
      id: 'ID',
      manager: 'WhiteListManager',
      vault: 'Vault',
      addresses: 'String',
      timestamp: 'BigInt',
      updatedTimestamp: 'BigInt'
    });
    this._entityTypesMap.set('LeaderBoard', {
      id: 'ID',
      address: 'String',
      numStaticJobs: 'BigInt',
      numApps: 'BigInt',
      timestamp: 'BigInt'
    });
    this._entityTypesMap.set('VaultTransfer', {
      id: 'ID',
      from: 'String',
      to: 'String',
      value: 'BigInt',
      timestamp: 'BigInt',
      blockNumber: 'BigInt'
    });
  }

  // eslint-disable-next-line @typescript-eslint/no-empty-function
  _populateRelationsMap (): void {
    this._relationsMap.set(Keeper, {
      actions: {
        entity: OrchestratorAction,
        isArray: true,
        isDerived: true,
        field: 'from'
      },
      queueTimeline: {
        entity: QueueTimeline,
        isArray: true,
        isDerived: true,
        field: 'keeper'
      },
      permissionUpdates: {
        entity: PermissionUpdate,
        isArray: true,
        isDerived: true,
        field: 'keeper'
      }
    });
    this._relationsMap.set(QueueTimeline, {
      keeper: {
        entity: Keeper,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(PermissionUpdate, {
      keeper: {
        entity: Keeper,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(OrchestratorAction, {
      from: {
        entity: Keeper,
        isArray: false,
        isDerived: false
      },
      votes: {
        entity: Vote,
        isArray: true,
        isDerived: true,
        field: 'action'
      },
      actionFailed: {
        entity: ActionFailure,
        isArray: true,
        isDerived: true,
        field: 'action'
      }
    });
    this._relationsMap.set(Vote, {
      by: {
        entity: Keeper,
        isArray: false,
        isDerived: false
      },
      action: {
        entity: OrchestratorAction,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(ActionFailure, {
      action: {
        entity: OrchestratorAction,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(Creator, {
      strategies: {
        entity: Strategy,
        isArray: true,
        isDerived: true,
        field: 'creator'
      },
      withdrawals: {
        entity: CreatorWithdrawal,
        isArray: true,
        isDerived: true,
        field: 'creator'
      }
    });
    this._relationsMap.set(CreatorWithdrawal, {
      creator: {
        entity: Creator,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(Strategy, {
      vault: {
        entity: Vault,
        isArray: true,
        isDerived: true,
        field: 'strategyToken'
      },
      creator: {
        entity: Creator,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(Vault, {
      statusUpdates: {
        entity: VaultStateChanged,
        isArray: true,
        isDerived: true,
        field: 'vault'
      },
      strategyToken: {
        entity: Strategy,
        isArray: false,
        isDerived: false
      },
      snapshots: {
        entity: VaultSnapshot,
        isArray: true,
        isDerived: true,
        field: 'vaultAddress'
      },
      permissions: {
        entity: WhiteListVaultPermission,
        isArray: true,
        isDerived: true,
        field: 'vault'
      },
      positions: {
        entity: VaultPosition,
        isArray: true,
        isDerived: true,
        field: 'vault'
      },
      depositors: {
        entity: Depositor,
        isArray: true,
        isDerived: true,
        field: 'vault'
      },
      jobs: {
        entity: Job,
        isArray: true,
        isDerived: true,
        field: 'vaultAddress'
      }
    });
    this._relationsMap.set(VaultSnapshot, {
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(PrevVaultSnapshot, {
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(PrevDailyVaultSnapshot, {
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(PrevMonthlyVaultSnapshot, {
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(PrevAnnualVaultSnapshot, {
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(PrevWeeklyVaultSnapshot, {
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(VaultStateChanged, {
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(VaultDeposit, {
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(VaultWithdraw, {
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(Depositor, {
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(VaultPosition, {
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(LiquiditySteer, {
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(OrchestratorReward, {
      trackerId: {
        entity: OrchestratorRewardTracker,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(OrchestratorRewardTracker, {
      rewards: {
        entity: OrchestratorReward,
        isArray: true,
        isDerived: true,
        field: 'trackerId'
      }
    });
    this._relationsMap.set(Job, {
      executions: {
        entity: JobExecution,
        isArray: true,
        isDerived: true,
        field: 'jobId'
      },
      vaultAddress: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(JobExecution, {
      jobId: {
        entity: Job,
        isArray: false,
        isDerived: false
      }
    });
    this._relationsMap.set(WhiteListManager, {
      permission: {
        entity: WhiteListVaultPermission,
        isArray: false,
        isDerived: true,
        field: 'manager'
      }
    });
    this._relationsMap.set(WhiteListVaultPermission, {
      manager: {
        entity: WhiteListManager,
        isArray: false,
        isDerived: false
      },
      vault: {
        entity: Vault,
        isArray: false,
        isDerived: false
      }
    });
  }

  async _saveBlockAndFetchEvents ({
    cid: blockCid,
    blockHash,
    blockNumber,
    blockTimestamp,
    parentHash
  }: DeepPartial<BlockProgress>): Promise<[
    BlockProgress,
    DeepPartial<Event>[],
    EthFullTransaction[]
  ]> {
    assert(blockHash);
    assert(blockNumber);

    let dbEvents: DeepPartial<Event>[] = [];
    let transactions: EthFullTransaction[] = [];

    // Fetch events and txs only if subgraph config has any event handlers
    if (this._graphWatcher.eventHandlerExists) {
      ({ events: dbEvents, transactions } = await this._baseIndexer.fetchEvents(
        blockHash,
        blockNumber,
        this.eventSignaturesMap,
        this.parseEventNameAndArgs.bind(this)
      ));
    }

    const dbTx = await this._db.createTransactionRunner();
    try {
      const block = {
        cid: blockCid,
        blockHash,
        blockNumber,
        blockTimestamp,
        parentHash
      };

      console.time(`time:indexer#_saveBlockAndFetchEvents-db-save-${blockNumber}`);
      const blockProgress = await this._db.saveBlockWithEvents(dbTx, block, dbEvents);
      await dbTx.commitTransaction();
      console.timeEnd(`time:indexer#_saveBlockAndFetchEvents-db-save-${blockNumber}`);

      return [blockProgress, [], transactions];
    } catch (error) {
      await dbTx.rollbackTransaction();
      throw error;
    } finally {
      await dbTx.release();
    }
  }

  async getFullTransactions (txHashList: string[]): Promise<EthFullTransaction[]> {
    return this._baseIndexer.getFullTransactions(txHashList);
  }
}
